require 'monad/base'
require 'ostruct'

module Rammed
  module Monad
    module EitherConstructors
      class Right
        include Monad::Base
        def initialize(v)
          @v = v
        end

        def right?
          true
        end

        def left?
          false
        end

        def bind(f)
          f.(@v)
        end

        def ==(other)
          other.respond_to?(:right?) && other.right? && other.instance_variable_get(:@v) == @v
        end

        def mreturn(v)
          self.class.new(v)
        end

        def extract
          @v
        end

        def assign(_, r)
          r.(@v)
          nil
        end
      end

      class Left
        include Monad::Base
        def initialize(v)
          @v = v
        end

        def right?
          false
        end

        def left?
          true
        end

        def bind(_)
          self
        end

        def ==(other)
          other.respond_to?(:left?) && other.left? && other.instance_variable_get(:@v) == @v
        end

        def extract
          @v
        end

        def assign(l, _)
          l.(@v)
          nil
        end
      end
    end

    Either = OpenStruct.new(COMMON_FUNCS.merge(
      of: ->(v) { v.nil? ? EitherConstructors::Left.new(nil) : EitherConstructors::Right.new(v) },
      right?: Rammed.invoker(0, :right?),
      left?: Rammed.invoker(0, :left?),
      right: EitherConstructors::Right.method(:new),
      left: EitherConstructors::Left.method(:new),
      extract: Rammed.invoker(0, :extract),
      assign: Rammed.invoker(2, :assign)
    ))
  end
end
