require 'monad/base'
require 'ostruct'

module Rammed
  module Monad
    module MaybeConstructors
      class Just
        include Monad::Base
        def initialize(v)
          @v = v
        end

        def just?
          true
        end

        def none?
          false
        end

        def bind(f)
          f.(@v)
        end

        def map(f)
          self >> ->(x) { Just.new(f.(x)) }
        end

        def ==(other)
          other.respond_to?(:just?) && other.just? && other.instance_variable_get(:@v) == @v
        end

        def mreturn(v)
          self.class.new(v)
        end
      end

      class None
        include Monad::Base
        def just?
          false
        end

        def none?
          true
        end

        def bind(_)
          self
        end

        def ==(other)
          other.respond_to?(:none?) && other.none?
        end
      end
    end
    Maybe = OpenStruct.new(COMMON_FUNCS.merge(
      of: ->(v) { v.nil? ? MaybeConstructors::None.new : MaybeConstructors::Just.new(v) },
      none?: Core.invoker.(0, :none?),
      just?: Core.invoker.(0, :just?),
      none: MaybeConstructors::None.method(:new),
      just: MaybeConstructors::Just.method(:new)
    ))
  end
end
